#!/usr/bin/python3

#  Copyright 2019 Matt Hart <matt@mattface.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import asyncio
import logging
logger = logging.getLogger('pdud.listener')


class Args(object):
    hostname = None
    alias = None
    request = None
    port = None
    pass


def parse_tcp(data):
    args = Args()
    delay = None
    array = data.split(" ")
    if (len(array) < 3) or (len(array) > 4):
        logger.info("Wrong data size")
        raise Exception("Unexpected data")
    if len(array) == 4:
        delay = int(array[3])
    args.hostname = array[0]
    args.port = int(array[1])
    args.request = array[2]
    args.delay = delay
    return args


def parse_http(data, path):
    args = Args()
    delay = None
    entry = path.lstrip('/').split('/')
    if len(entry) != 3:
        logger.info("Request path was invalid: %s", entry)
        return False
    if not (entry[0] == 'power' and entry[1] == 'control'):
        logger.info("Unknown request, path was %s", path)
        return False
    # everything comes back from the http library in a 1 sized list
    args.alias = data.get('alias', [None])[0]
    args.hostname = data.get('hostname', [None])[0]
    args.port = data.get('port', [None])[0]
    args.request = entry[2]
    args.delay = data.get('delay', [None])[0]
    return args


async def process_request(args, config, daemon):
    if args.request in ["on", "off"] and args.delay is not None:
        logger.warn("delay parameter is deprecated for on/off commands")
    if args.delay is not None:
        logger.debug("using custom delay as requested")
    else:
        # this has been a default since the start, it should be smaller but I
        # can't really change expected behaviour now
        if args.request == "reboot":
            args.delay = 5
        else:
            args.delay = 0
    if args.alias:
        if args.hostname or args.port:
            logging.error("Trying to use and alias and also a hostname/port")
            return False
        # Using alias support, get all pdu info from alias
        alias_settings = (config.get('aliases', {})).get(args.alias, False)
        if not alias_settings:
            logging.error("Alias requested but not found")
            return False
        args.hostname = config["aliases"][args.alias]["hostname"]
        args.port = config["aliases"][args.alias]["port"]
    if not args.hostname or not args.port or not args.request:
        logger.info("One of hostname,port,request was not set")
        return False
    if args.hostname not in config['pdus']:
        logger.info("PDU was not found in config")
        return False
    if not (args.request in ["reboot", "on", "off"]):
        logger.info("Unknown request: %s", args.request)
        return False
    runner = daemon.runners[args.hostname]
    if args.request == "reboot":
        logger.debug("reboot requested, submitting off/on")
        await runner.do_job_async(int(args.port), "off")
        await asyncio.sleep(int(args.delay))
        await runner.do_job_async(int(args.port), "on")
        return True
    else:
        await asyncio.sleep(int(args.delay))
        await runner.do_job_async(int(args.port), args.request)
        return True



#!/usr/bin/python3
#
#  Copyright 2020 Joshua Watt <JPEWhacker@gmail.com>
#
#  A driver for V-USB HID based low cost relays
#
#  See: http://vusb.wikidot.com/project:driver-less-usb-relays-hid-interface
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import logging
from pdudaemon.drivers.driver import PDUDriver
import usb.core
import usb.util
import array

import os

log = logging.getLogger("pdud.drivers." + os.path.basename(__file__))

VENDOR_ID = 0x16C0
PRODUCT_ID = 0x05DF
USB_TYPE_CLASS = 0x20
USB_ENDPOINT_OUT = 0x00
USB_ENDPOINT_IN = 0x80
USB_RECIP_DEVICE = 0x00
GET_REPORT = 0x1
SET_REPORT = 0x9
USB_HID_REPORT_TYPE_FEATURE = 3
MAIN_REPORT = 0


class VUSBHID(PDUDriver):
    def __init__(self, hostname, settings):
        self.settings = settings
        self.serial = settings.get("serial", "")
        self.invert = settings.get("invert", False)
        log.debug("serial: %s" % self.serial)

    def port_interaction(self, command, port_number):
        def xor(a, b):
            return bool(a) != bool(b)

        devices = self.connect()

        if len(devices) == 0:
            err = "No device found"
            log.error(err)
            raise RuntimeError(err)

        dev = None
        for d in devices:
            product = usb.util.get_string(d, d.iProduct)
            if not product.startswith("USBRelay"):
                continue

            data = self.get_report(d, MAIN_REPORT, 8)
            serial = data[0:5].tobytes().decode("utf-8")

            if serial == self.serial:
                dev = d
                break
        else:
            err = "Device with serial {} not found".format(self.serial)
            log.error(err)
            raise RuntimeError(err)

        port_number = int(port_number)

        # The end of the product string indicates the number of relays
        max_ports = int(product[8:])

        if port_number > max_ports or port_number < 1:
            err = "Port should be in the range 1 - {}".format(max_ports)
            log.error(err)
            raise RuntimeError(err)

        if command == "on":
            self.set_state(dev, port_number, xor(True, self.invert))
        elif command == "off":
            self.set_state(dev, port_number, xor(False, self.invert))
        else:
            log.error("Unknown command %s." % (command))
            return

    @classmethod
    def accepts(cls, drivername):
        return drivername == "vusbhid"

    def connect(self):
        ret = list()
        ret += list(usb.core.find(find_all=True, idVendor=0x16C0, idProduct=0x05DF))
        return ret

    def set_state(self, dev, port, state):
        buf = array.array("B")
        buf.append(0xFF if state else 0xFD)
        buf.append(port)
        dev.ctrl_transfer(
            USB_TYPE_CLASS | USB_RECIP_DEVICE | USB_ENDPOINT_OUT,
            SET_REPORT,
            (USB_HID_REPORT_TYPE_FEATURE << 8) | MAIN_REPORT,
            0,
            buf,
            5000,
        )

    def get_report(self, dev, report, size):
        return dev.ctrl_transfer(
            USB_TYPE_CLASS | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
            GET_REPORT,
            (USB_HID_REPORT_TYPE_FEATURE << 8) | report,
            0,
            size,
            5000,
        )



#!/usr/bin/python3

#  Copyright 2018 Sjoerd Simons <sjoerd.simons@collabora.co.uk>
#                 Stefan Kempe <stefan.kempe@de.bosch.com>
#
#  Based on PDUDriver:
#     Copyright 2013 Linaro Limited
#     Author Matt Hart <matthew.hart@linaro.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import logging
from pdudaemon.drivers.driver import PDUDriver
import hid
import os

log = logging.getLogger("pdud.drivers." + os.path.basename(__file__))

YKUSH_VID = 0x04d8
YKUSH_PID = 0xf2f7
YKUSH_XS_PID = 0xf0cd
YKUSH3_PID = 0xf11b


class YkushBase(PDUDriver):
    connection = None
    ykush_pid = None
    port_count = 0

    def __init__(self, hostname, settings):
        self.hostname = hostname
        self.settings = settings
        self.serial = settings.get("serial", u"")
        log.debug("serial: %s" % self.serial)

        super().__init__()

    def port_interaction(self, command, port_number):
        port_number = int(port_number)
        if port_number > self.port_count or port_number < 1:
            err = "Port should be in the range 1 - %d" % (self.port_count)
            log.error(err)
            raise RuntimeError(err)

        if command == "on":
            byte = 0x10 + port_number
        elif command == "off":
            byte = 0x00 + port_number
        else:
            log.error("Unknown command %s." % (command))
            return

        d = hid.device()
        d.open(YKUSH_VID, self.ykush_pid, serial_number=self.serial)
        d.write([byte, byte])
        d.read(64)
        d.close()

    @classmethod
    def accepts(cls, drivername):
        return drivername == cls.__name__.upper()


class YkushXS(YkushBase):
    ykush_pid = YKUSH_XS_PID
    port_count = 1


class Ykush(YkushBase):
    ykush_pid = YKUSH_PID
    port_count = 3


class Ykush3(YkushBase):
    ykush_pid = YKUSH3_PID
    port_count = 3



#!/usr/bin/python3

#  Copyright 2013 Linaro Limited
#  Author Matt Hart <matthew.hart@linaro.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import logging
import pexpect
from pdudaemon.drivers.driver import PDUDriver
import sys
import os
log = logging.getLogger("pdud.drivers." + os.path.basename(__file__))


class LocalBase(PDUDriver):
    connection = None

    def __init__(self, hostname, settings):
        self.hostname = hostname
        self.settings = settings
        super(LocalBase, self).__init__()

    @classmethod
    def accepts(cls, drivername):
        return False

    def port_interaction(self, command, port_number):
        log.debug("Running port_interaction from LocalBase")
        log.debug("Attempting control: {} port: {}".format(command, port_number))
        self._port_interaction(command, port_number)

    def _bombout(self):
        log.debug("Bombing out of driver: %s" % self.connection)
        del (self)

    def _cleanup(self):
        log.debug("Cleaning up driver: %s" % self.connection)



#!/usr/bin/python3

#  Copyright 2016 Broadcom
#  Author Christian Daudt <csd@broadcom.com>
#  Based on apcbase+apc8959 by:
#  Author Matt Hart <matthew.hart@linaro.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import sys
import logging
import pexpect
from pdudaemon.drivers.driver import PDUDriver
import os
log = logging.getLogger("pdud.drivers." + os.path.basename(__file__))


class SynBase(PDUDriver):
    connection = None

    def __init__(self, hostname, settings):
        self.hostname = hostname
        log.debug(settings)
        self.settings = settings
        self.username = "admin"
        self.password = "admin"
        telnetport = 23

        if "telnetport" in settings:
            telnetport = settings["telnetport"]
        if "username" in settings:
            self.username = settings["username"]
        if "password" in settings:
            self.password = settings["password"]

        self.exec_string = "/usr/bin/telnet %s %d" % (hostname, telnetport)
        log.debug("Telnet command: [%s]" % self.exec_string)
        super(SynBase, self).__init__()

    @classmethod
    def accepts(cls, drivername):
        return False

    def port_interaction(self, command, port_number):
        log.debug("Running port_interaction from SynBase")
        self.get_connection()
        log.debug("Attempting {} on port {}".format(command, port_number))
        self._port_interaction(command,  # pylint: disable=no-member
                               port_number)

    def get_connection(self):
        log.debug("Connecting to Syn PDU with: %s", self.exec_string)
        # only uncomment this line for FULL debug when developing
        # self.connection = pexpect.spawn(self.exec_string, logfile=sys.stdout)
        self.connection = pexpect.spawn(self.exec_string)
        self._pdu_login(self.username, self.password)

    def _cleanup(self):
        self._pdu_logout()  # pylint: disable=no-member

    def _bombout(self):
        log.debug("Bombing out of driver: %s", self.connection)
        self.connection.close(force=True)
        del self

    def _pdu_login(self, username, password):
        # Expected sequence:
        # >login
        # User ID: admin
        # Password:******
        # >

        log.debug("attempting login with username %s, password %s",
                  username, password)
        self.connection.expect(">")
        self.connection.send("login\r")
        self.connection.expect("User ID:")
        self.connection.send("%s\r" % username)
        self.connection.expect("Password:")
        self.connection.send("%s\r" % password)
        self.connection.expect(">")

#
# Only Synaccess product support at this point
# Synaccess Networks netBooter Series B
# Login identifies as: System Model: NP-08B


class SynNetBooter(SynBase):
    pdu_commands = {"off": "pset %s 0", "on": "pset %s 1"}

    @classmethod
    def accepts(cls, drivername):
        if drivername == "synnetbooter":
            return True
        return False

    def _pdu_logout(self):
        log.debug("logging out")
        self.connection.send("\r")
        self.connection.send("exit")
        self.connection.send("\r")
        log.debug("done")

    def _pdu_get_to_prompt(self):
        self.connection.send("\r")
        self.connection.expect('>')

    def _port_interaction(self, command, port_number):
        self._pdu_get_to_prompt()
        self.connection.sendline(self.pdu_commands[command] %
                                 (port_number))
        self._pdu_get_to_prompt()
        log.debug("done")



#!/usr/bin/python3

#  Copyright 2013 Linaro Limited
#  Author Matt Hart <matthew.hart@linaro.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import asyncio
import logging
import socket
import time
import pdudaemon.listener as listener
logger = logging.getLogger('pdud.tcp')


class TCPListener:

    def __init__(self, config, daemon):
        self.config = config
        self.daemon = daemon
        self.settings = config["daemon"]

        self.server = None

    async def start(self):
        listen_host = self.settings["hostname"]
        listen_port = self.settings.get("port", 16421)
        logger.info("listening on %s:%s", listen_host, listen_port)
        self.server = await asyncio.start_server(
            client_connected_cb=self.handle,
            host=listen_host,
            port=listen_port,
            reuse_address=True,
        )

    async def shutdown(self):
        if self.server:
            self.server.close()
            await self.server.wait_closed()
        self.server = None

    async def insert_request(self, data):
        args = listener.parse_tcp(data)
        if args:
            return await listener.process_request(args, self.config, self.daemon)

    async def handle(self, reader, writer):
        request_ip = writer.get_extra_info('peername')[0]
        loop = asyncio.get_running_loop()
        try:
            data = await reader.read(16384)
            data = data.decode('utf-8')
            data = data.strip()
            try:
                fut = loop.run_in_executor(None, socket.gethostbyaddr, request_ip)
                request_host = (await asyncio.wait_for(fut, timeout=2))[0]
            except (socket.herror, asyncio.TimeoutError):
                request_host = request_ip
            logger.info("Received a request from %s: '%s'", request_host, data)
            res = await self.insert_request(data)
            if res:
                writer.write("ack\n".encode('utf-8'))
            else:
                writer.write("nack\n".encode('utf-8'))
        except Exception as global_error:  # pylint: disable=broad-except
            logger.debug(global_error.__class__)
            logger.debug(global_error)
            writer.write(str(global_error).encode('utf-8'))
        writer.close()
        await writer.wait_closed()



#!/usr/bin/python3

# It turns out that the device ID of the EnerGenie devices isn't the same
# as the serial number printed on the unit. This script will return the
# IDs of the devices found connected. The ID needs to be used in the
# "device" field when configuring pdudaemon.conf to use an energenie PDU.

import usb.core

dev = list()
dev += list(usb.core.find(find_all=True, idVendor=0x04b4, idProduct=0xfd10))
dev += list(usb.core.find(find_all=True, idVendor=0x04b4, idProduct=0xfd11))
dev += list(usb.core.find(find_all=True, idVendor=0x04b4, idProduct=0xfd12))
dev += list(usb.core.find(find_all=True, idVendor=0x04b4, idProduct=0xfd13))
dev += list(usb.core.find(find_all=True, idVendor=0x04b4, idProduct=0xfd15))

for d in dev:
    buf = bytes([0x00, 0x00, 0x00, 0x00, 0x00])
    ret = d.ctrl_transfer(0xa1, 0x01, 0x0301, 0, buf, 500)
    id = ":".join(format(x, '02x') for x in ret.tolist())
    print(id)



#!/usr/bin/python3
#
#  Copyright 2018 Remi Duraffort <remi.duraffort@linaro.org>
#                 Matt Hart <matt@mattface.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

__all__ = ["__author__", "__description__", "__license__", "__url__",
           "__version__"]


def git_hash():
    import subprocess
    try:
        out = subprocess.check_output(["git", "rev-parse", "--short", "HEAD"],
                                      stderr=subprocess.STDOUT)
        return out.decode("utf-8").rstrip("\n")
    except Exception:
        return "git"


__author__ = "Matt Hart"
__description__ = 'Control and Queueing daemon for PDUs'
__license__ = 'GPLv2+'
__url__ = 'https://github.com/pdudaemon/pdudaemon.git'
__version__ = "0.1"



#!/usr/bin/python3

#
#  Copyright 2016 BayLibre, Inc.
#  Author Kevin Hilman <khilman@kernel.org>
#
#  Based on localcmdline.py
#  Copyright 2013 Linaro Limited
#  Author Matt Hart <matthew.hart@linaro.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import logging
from pdudaemon.drivers.localbase import LocalBase
import requests

import os
log = logging.getLogger("pdud.drivers." + os.path.basename(__file__))


class Sainsmart(LocalBase):

    def __init__(self, hostname, settings):
        self.hostname = hostname
        self.settings = settings
        self.ip = settings.get("ip", self.hostname)
        self.url_base = "http://%s/30000/" % self.ip
        log.debug(self.url_base)

    @classmethod
    def accepts(cls, drivername):
        if drivername == "sainsmart":
            return True
        return False

    def _port_interaction(self, command, port_number):

        val = -1
        if command == "on":
            log.debug("Attempting control: {} port: {}".format(command, port_number))
            val = (port_number - 1) * 2 + 1

        elif command == "off":
            log.debug("Attempting control: {} port: {}".format(command, port_number))
            val = (port_number - 1) * 2

        else:
            log.debug("Unknown command!")

        if (val >= 0):
            url = self.url_base + "%02d" % val
            log.debug("HTTP GET at %s" % url)
            requests.get(url)



#!/usr/bin/python3

#  Copyright 2013 Linaro Limited
#  Author Matt Hart <matthew.hart@linaro.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import logging
from pdudaemon.drivers.apcbase import APCBase
import os
log = logging.getLogger("pdud.drivers." + os.path.basename(__file__))


class APC8959(APCBase):
    pdu_commands = {"off": "olOff", "on": "olOn"}

    @classmethod
    def accepts(cls, drivername):
        if drivername == "apc8959":
            return True
        return False

    def _pdu_logout(self):
        log.debug("logging out")
        self.connection.send("\r")
        self.connection.send("exit")
        self.connection.send("\r")
        log.debug("done")

    def _pdu_get_to_prompt(self):
        self.connection.send("\r")
        self.connection.expect('apc>')

    def _port_interaction(self, command, port_number):
        self._pdu_get_to_prompt()
        self.connection.sendline(self.pdu_commands[command] + (" {}".format(port_number)))
        self.connection.expect("E000: Success")
        self._pdu_get_to_prompt()
        log.debug("done")



import pytest
from unittest import mock
from .numatousb import NumatoUSB4
from .numatousb import NumatoUSB32
from .numatousb import NumatoUSB64


def test_numato4():
    with mock.patch('serial.serial_for_url') as m:
        dev = NumatoUSB4('h', dict(device='/dev/fake'))
        dev.port_on(3)
    m.return_value.write.assert_called_with(b'relay on 2\r')


def test_numato32():
    with mock.patch('serial.serial_for_url') as m:
        dev = NumatoUSB32('h', dict(device='/dev/fake'))
        dev.port_on(32)
    m.return_value.write.assert_called_with(b'relay on V\r')


def test_numato64():
    with mock.patch('serial.serial_for_url') as m:
        dev = NumatoUSB64('h', dict(device='/dev/fake'))
        dev.port_on(32)
    m.return_value.write.assert_called_with(b'relay on 31\r')



#!/usr/bin/python3

#  Copyright 2022 Laura Nao <laura.nao@collabora.com>
#
#  Based on PDUDriver:
#     Copyright 2013 Linaro Limited
#     Author Matt Hart <matthew.hart@linaro.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import logging
import os
from pdudaemon.drivers.driver import PDUDriver
from xmlrpc import client

log = logging.getLogger("pdud.drivers." + os.path.basename(__file__))


class Servo(PDUDriver):
    supported_ctrls = {'power_state': 'active_high',
                       'warm_reset': 'active_low',
                       'cold_reset': 'active_low'}

    def __init__(self, hostname, settings):
        self.hostname = hostname
        self.settings = settings

        self.ctrls = settings.get("ctrls", ["cold_reset"])
        if not isinstance(self.ctrls, list):
            self.ctrls = [self.ctrls]

        ip = settings.get("ip", "localhost")
        port = settings.get("port", 9999)
        self.remote_uri = "http://{}:{}".format(ip, port)

        log.debug("Servod server: %s" % self.remote_uri)

        super().__init__()

    def port_interaction(self, command, port_number):
        if command not in ('on', 'off'):
            log.error("Unknown command %s." % (command))
            return

        for ctrl in self.ctrls:
            if ctrl not in self.supported_ctrls:
                err = (f'Unknown control {ctrl}. '
                       f'Available controls: {list(self.supported_ctrls.keys())}')
                raise ValueError(err)

        for ctrl in self.ctrls:
            if self.supported_ctrls[ctrl] == 'active_low':
                val = 'on' if command == 'off' else 'off'
            else:
                val = command

            log.debug("Setting %s:%s" % (ctrl, val))

            with client.ServerProxy(self.remote_uri) as proxy:
                proxy.set(ctrl, val)

    @classmethod
    def accepts(cls, drivername):
        return drivername == cls.__name__.upper()



#!/usr/bin/python3

#  Copyright 2015 BayLibre SAS
#  Author Marc Titinger <mtitinger@baylibre.com>
#
#  Based on apcxxx:
#     Copyright 2013 Linaro Limited
#     Author Matt Hart <matthew.hart@linaro.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import logging
from pdudaemon.drivers.acmebase import ACMEBase
import os
log = logging.getLogger("pdud.drivers." + os.path.basename(__file__))


class ACME(ACMEBase):

    cmd = {'on': 'dut-switch-on', 'off': 'dut-switch-off', 'reboot': 'dut-hard-reset'}

    @classmethod
    def accepts(cls, drivername):
        if drivername == "acme":
            return True
        return False

    def _pdu_logout(self):
        self._back_to_main()
        log.debug("Logging out")
        self.connection.send("exit\r")

    def _back_to_main(self):
        self.connection.send("\r")
        self.connection.expect('#')

    def _enter_outlet(self, outlet, enter_needed=True):
        log.debug("Finished entering outlet (nop)")

    def _port_interaction(self, command, port_number):
        if command not in self.cmd:
            acme_command = 'echo "unknown command {}"'.format(command)
        else:
            acme_command = '{} {}'.format(self.cmd[command], port_number)
        log.debug("Attempting command: %s", acme_command)
        self.connection.send(acme_command)
        self._do_it()
        self.connection.expect("#")

    def _do_it(self):
        self.connection.send("\r")



#!/usr/bin/python3

#
#  Copyright 2021 Mark Ferry <github@markferry.net>
#
#  Based on PDUDriver:
#     Copyright 2013 Linaro Limited
#     Author Matt Hart <matthew.hart@linaro.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import logging
from pdudaemon.drivers.driver import PDUDriver, FailedRequestException
import requests
import os

log = logging.getLogger("pdud.drivers." + os.path.basename(__file__))

__manual__ = "https://s3.amazonaws.com/assets.mhint/downloads/61413/INT_163682_UM_0819_REV_5.03.pdf"


class Intellinet(PDUDriver):
    """
    Intellinet 163682 support

    See also https://github.com/01programs/Intellinet_163682_IP_smart_PDU_API
    """

    port_count = 8
    endpoints = {
        "outlet": "control_outlet.htm",
    }

    def __init__(self, hostname, settings):
        self.hostname = hostname
        self.port = settings.get("port", 80)
        self.username = settings.get("username", "admin")
        self.password = settings.get("password", "admin")

        self.auth = self._auth(self.username, self.password)

        super().__init__()

    def _auth(self, username, password):
        return requests.auth.HTTPBasicAuth(username, password)

    def _api(self, page, params):
        url = "http://{}:{}/{}".format(self.hostname, self.port, page)

        log.debug("HTTP GET: {}".format(url))
        r = requests.get(url, params, auth=self.auth)

        r.raise_for_status()
        if r.text.find("401") != -1:
            log.error(r.text)
            raise FailedRequestException(r.text)
        log.debug("HTTP response: {}".format(r.text))

    def port_interaction(self, command, port_number):
        port_number = int(port_number)

        if port_number >= self.port_count or port_number < 0:
            err = "Port number must be in range 0 - {}".format(self.port_count - 1)
            log.error(err)
            raise FailedRequestException(err)

        command_map = {"on": 0, "off": 1, "reboot": 2}

        if command not in command_map:
            log.error("Unknown command %s." % (command))
            raise FailedRequestException(err)

        endpoint = self.endpoints['outlet']
        params = {"outlet{}".format(port_number): 1}
        params["op"] = command_map[command]
        params["submit"] = "Anwenden"
        self._api(endpoint, params)

    @classmethod
    def accepts(cls, drivername):
        if drivername == "intellinet":
            return True
        return False



#!/usr/bin/python3

#  Copyright 2020
#  Author Matt Hart <matt@mattface.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Confirmed devices
#    - KP303: https://www.tp-link.com/uk/home-networking/smart-plug/kp303/
#    - KP105: https://www.tp-link.com/uk/home-networking/smart-plug/kp105/
#    - HS105: https://www.tp-link.com/us/home-networking/smart-plug/hs105/

import logging
import json
import socket
from pdudaemon.drivers.driver import PDUDriver
from struct import pack
import os
log = logging.getLogger("pdud.drivers." + os.path.basename(__file__))


class TPLink(PDUDriver):
    connection = None

    def __init__(self, hostname, settings):
        self.hostname = hostname
        self.settings = settings
        self.childinfo = {}
        self.getinfo()
        super(TPLink, self).__init__()

    @classmethod
    def accepts(cls, drivername):
        if drivername == "tplink":
            return True
        return False

    def encrypt(self, string):
        key = 171
        result = pack(">I", len(string))
        for i in string:
            a = key ^ ord(i)
            key = a
            result += bytes([a])
        return result

    def decrypt(self, string):
        key = 171
        result = ""
        for i in string:
            a = key ^ i
            key = i
            result += chr(a)
        return result

    def getinfo(self):
        datadict = {
            'system': {
                'get_sysinfo': {
                }
            }
        }
        res = self.send_command(json.dumps(datadict))
        if res:
            resdict = json.loads(res)
            self.childinfo = resdict.get("system", {}).get("get_sysinfo", {}).get("children", {})

    def send_command(self, json_string):
        try:
            sock_tcp = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock_tcp.settimeout(10)
            sock_tcp.connect((self.hostname, 9999))
            sock_tcp.settimeout(None)
            sock_tcp.send(self.encrypt(json_string))
            data = sock_tcp.recv(2048)
            sock_tcp.close()
            decrypted = self.decrypt(data[4:])
            return (decrypted)
        except socket.error:
            log.error(f"Could not connect to host {self.hostname}:9999")

    def get_context(self, port_number):
        for child in self.childinfo:
            child_port = (int(child['alias'].split("_")[-1]) + 1)
            if int(port_number) == int(child_port):
                return ({"child_ids": [child["id"]]})
        return ({})

    def port_interaction(self, command, port_number):
        state = 0
        context = None
        if command == "on":
            state = 1
        if self.childinfo:
            if int(port_number) > len(self.childinfo):
                return (False)
            context = self.get_context(port_number)
        datadict = {
            'context': context,
            'system': {
                'set_relay_state': {
                    'state': state,
                }
            }
        }
        log.debug(datadict)
        res = self.send_command(json.dumps(datadict))



from unittest.mock import ANY

import pytest
from pdudaemon.drivers.driver import FailedRequestException
from pdudaemon.drivers.intellinet import Intellinet


@pytest.fixture
def pdu():
    return Intellinet("dummy", {})


@pytest.fixture(name="api_mock")
def fixture_api_mock(mocker):
    return mocker.patch("pdudaemon.drivers.intellinet.Intellinet._api")


def test_port_interaction(pdu, api_mock):
    pdu.port_interaction("on", 0)
    api_mock.assert_called_once_with(pdu.endpoints["outlet"], ANY)


def test_port_interaction_port_in_range(pdu):
    pdu = Intellinet("dummy", {})

    with pytest.raises(FailedRequestException):
        pdu.port_interaction("on", pdu.port_count)

    with pytest.raises(FailedRequestException):
        pdu.port_interaction("on", -1)



#!/usr/bin/python3

#  Copyright 2018 Matt Hart <matt@mattface.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import urllib.parse as urlparse
import logging
import pdudaemon.listener as listener

from aiohttp import web

logger = logging.getLogger('pdud.http')


class HTTPListener:

    def __init__(self, config, daemon):
        self.config = config
        self.daemon = daemon
        self.settings = config["daemon"]

        self.app = web.Application()
        self.app.add_routes([
            web.get('/power/control/on', self.handle),
            web.get('/power/control/off', self.handle),
            web.get('/power/control/reboot', self.handle),
        ])
        self.apprunner = None

    async def start(self):
        logger.info("Starting the HTTP server")
        self.apprunner = web.AppRunner(self.app)
        await self.apprunner.setup()
        listen_host = self.settings["hostname"]
        listen_port = self.settings.get("port", 16421)
        site = web.TCPSite(self.apprunner, host=listen_host, port=listen_port)
        await site.start()
        logger.info("Listening on %s:%s", listen_host, listen_port)

    async def shutdown(self):
        if self.apprunner:
            await self.apprunner.cleanup()
        self.apprunner = None

    async def handle(self, request):
        logger.info("Handling HTTP request from %s: %s", request.remote, request.path_qs)
        data = urlparse.parse_qs(urlparse.urlparse(request.path_qs).query)
        path = urlparse.urlparse(request.path_qs).path
        res = await self.insert_request(data, path)
        if res:
            return web.Response(status=200, text="OK - accepted request\n")
        else:
            return web.Response(status=500, text="Invalid request\n")

    async def insert_request(self, data, path):
        args = listener.parse_http(data, path)
        if args:
            return await listener.process_request(args, self.config, self.daemon)



#!/usr/bin/python3

#  Copyright 2019 Stefan Wiehler <stefan.wiehler@missinglinkelectronics.com>
#
#  Based on PDUDriver:
#     Copyright 2013 Linaro Limited
#     Author Matt Hart <matthew.hart@linaro.org>
#
#  Protocol documentation (in German) available at:
#  https://forum.anel.eu/viewtopic.php?f=52&t=888&sid=a0aca2195ffff4eb28a11fd89898590b
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import logging
from pdudaemon.drivers.driver import PDUDriver, FailedRequestException
import requests
import os
log = logging.getLogger("pdud.drivers." + os.path.basename(__file__))


class AnelNETPwrCtrlBase(PDUDriver):
    def __init__(self, hostname, settings):
        self.hostname = hostname
        self.username = settings.get("username", "admin")
        self.password = settings.get("password", "anel")
        super().__init__()

    def port_interaction(self, command, port_number):
        if port_number > self.port_count or port_number < 1:
            err = "Port number must be in range 1 - {}".format(self.port_count)
            log.error(err)
            raise FailedRequestException(err)

        url = "http://{}?Sw_{}={},{}{}".format(
            self.hostname, command, port_number, self.username, self.password)
        log.debug("HTTP GET: {}".format(url))
        r = requests.get(url)

        r.raise_for_status()
        if r.text == "User or password error.":
            log.error(r.text)
            raise FailedRequestException(r.text)
        log.debug('HTTP response: {}'.format(r.text))

    @classmethod
    def accepts(cls, drivername):
        return False


class AnelNETPwrCtrlHOME(AnelNETPwrCtrlBase):
    port_count = 3

    @classmethod
    def accepts(cls, drivername):
        if drivername == "anel_netpwrctrlhome":
            return True
        return False


class AnelNETPwrCtrlADV(AnelNETPwrCtrlBase):
    port_count = 8

    @classmethod
    def accepts(cls, drivername):
        if drivername == "anel_netpwrctrladv":
            return True
        return False


class AnelNETPwrCtrlIO(AnelNETPwrCtrlBase):
    port_count = 8

    @classmethod
    def accepts(cls, drivername):
        if drivername == "anel_netpwrctrlio":
            return True
        return False


class AnelNETPwrCtrlPRO(AnelNETPwrCtrlBase):
    port_count = 8

    @classmethod
    def accepts(cls, drivername):
        if drivername == "anel_netpwrctrlpro":
            return True
        return False



#!/usr/bin/python3
#
#  Copyright 2022 Christopher Obbard <chris.obbard@collabora.com>
#
#  Based on PDUDriver:
#     Copyright 2013 Linaro Limited
#     Author Matt Hart <matthew.hart@linaro.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import logging
import os
from pdudaemon.drivers.driver import PDUDriver, FailedRequestException
import requests
from requests.auth import HTTPDigestAuth

log = logging.getLogger("pdud.drivers." + os.path.basename(__file__))


class ESPHomeHTTP(PDUDriver):
    def __init__(self, hostname, settings):
        """Communicate with custom devices flashed with ESPHome firmware, specifically
        Switch components (e.g. relays, GPIO).
        To use this driver, the `web_server` stanza must be defined in the ESPHome
        configuration.
        """
        self.hostname = hostname
        self.port = settings.get("port", 80)
        self.username = settings.get("username")
        self.password = settings.get("password")

        self.switch_ids = settings.get("switch_ids")
        if self.switch_ids is None:
            raise RuntimeError(
                "No switch entity ID defined for %s. Provide `switch_ids` configuration entry with a list of switch IDs."
                % self.hostname
            )
        self.port_count = len(self.switch_ids)

        super().__init__()

    def port_interaction(self, command, port_number):
        esphome_cmd = ""
        if command == "on":
            esphome_cmd = "turn_on"
        elif command == "off":
            esphome_cmd = "turn_off"
        else:
            raise FailedRequestException("Unknown command %s" % (command))

        if int(port_number) > self.port_count or int(port_number) < 1:
            err = "Port number must be in range 1 - {}".format(self.port_count)
            log.error(err)
            raise FailedRequestException(err)
        esphome_entity_id = self.switch_ids[port_number - 1]

        # Build the POST request
        # url should be in the format http://{hostname}/switch/{id}/{cmd}
        url = "http://{}/switch/{}/{}".format(
            self.hostname, esphome_entity_id, esphome_cmd
        )
        log.debug("HTTP POST: {}".format(url))

        auth = None
        if self.username and self.password:
            auth = HTTPDigestAuth(self.username, self.password)
        response = requests.post(url, auth=auth)

        log.debug(
            "Response code for request to {}: {}".format(
                self.hostname, response.status_code
            )
        )
        response.raise_for_status()

    @classmethod
    def accepts(cls, drivername):
        return drivername == "esphome-http"



#!/usr/bin/python3

#  Copyright 2013 Linaro Limited
#  Author Matt Hart <matthew.hart@linaro.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import logging
import os
log = logging.getLogger("pdud.drivers")


def get_named_entry_point(group, name):
    import pkg_resources
    eps = list(pkg_resources.iter_entry_points(group, name))
    if len(eps) > 1:
        raise Exception('Multiple entry points for {} under {}'.format(group, name))
    if len(eps) == 0:
        return None
    return eps[0]


class PDUDriver(object):
    connection = None
    hostname = ""

    def __init__(self):
        super(PDUDriver, self).__init__()

    @classmethod
    def select(cls, drivername):
        ep = get_named_entry_point('pdudaemon.driver', drivername)
        if ep:
            # Not clear why a driver would reject the driver
            # it is registered for but check anyway:
            c = ep.load()
            if not c.accepts(drivername):
                raise Exception('pdudaemon.driver entry_point {} did not accept configuration'.format(c))
            return c
        candidates = cls.__subclasses__()  # pylint: disable=no-member
        for subc in cls.__subclasses__():  # pylint: disable=no-member
            candidates = candidates + (subc.__subclasses__())
            for subsubc in subc.__subclasses__():
                candidates = candidates + (subsubc.__subclasses__())
        willing = [c for c in candidates if c.accepts(drivername)]
        if len(willing) == 0:
            log.error("No driver accepted the configuration '%s'", drivername)
            os._exit(1)
        log.debug("%s accepted the request", willing[0])
        return willing[0]

    def handle(self, request, port_number):
        log.debug("Driving PDU hostname: %s "
                  "PORT: %s REQUEST: %s",
                  self.hostname, port_number, request)
        if request == "on":
            self.port_on(port_number)
        elif request == "off":
            self.port_off(port_number)
        else:
            log.debug("Unknown request to handle - oops")
            raise UnknownCommandException(
                "Driver doesn't know how to %s " % request
            )
        self._cleanup()

    def port_on(self, port_number):
        self.port_interaction("on", port_number)

    def port_off(self, port_number):
        self.port_interaction("off", port_number)

    def port_interaction(self, command, port_number):
        pass

    def _bombout(self):
        pass

    def _cleanup(self):
        pass


class UnknownCommandException(Exception):
    pass


class FailedRequestException(Exception):
    pass



#!/usr/bin/python3

#  Copyright 2016 Quentin Schulz <quentin.schulz@free-electrons.com>
#
#  Based on PDUDriver:
#     Copyright 2013 Linaro Limited
#     Author Matt Hart <matthew.hart@linaro.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import logging
from pdudaemon.drivers.driver import PDUDriver
import socket
import os
log = logging.getLogger("pdud.drivers." + os.path.basename(__file__))


class DevantechBase(PDUDriver):
    connection = None
    port_count = 0

    def __init__(self, hostname, settings):
        self.hostname = hostname
        self.settings = settings
        self.ip = settings["ip"]
        self.port = settings.get("port", 17494)
        self.password = settings.get("password")
        super(DevantechBase, self).__init__()

    def connect(self):
        self.connection = socket.create_connection((self.ip, self.port))
        if self.password:
            log.debug("Attempting connection to %s:%s with provided password.", self.hostname, self.port)
            msg = b'\x79' + self.password.encode("utf-8")
            ret = self.connection.sendall(msg)
            if ret:
                log.error("Failed to send message.")
                raise RuntimeError("Failed to send message.")
            ret = self.connection.recv(1)
            if ret != b'\x01':
                log.error("Authentication failed.")
                raise RuntimeError("Failed to authenticate. Verify your password.")

    def port_interaction(self, command, port_number):
        self.connect()
        port_number = int(port_number)
        if port_number > self.port_count:
            log.error("There are only %d ports. Provide a port number lesser than %d." % (self.port_count, self.port_count))
            raise RuntimeError("There are only %d ports. Provide a port number lesser than %d." % (self.port_count, self.port_count))

        if command == "on":
            msg = b'\x20'
        elif command == "off":
            msg = b'\x21'
        else:
            log.error("Unknown command %s." % (command))
            return
        msg += port_number.to_bytes(1, 'big')
        msg += b'\x00'
        log.debug("Attempting control: %s port: %d hostname: %s." % (command, port_number, self.hostname))
        ret = self.connection.sendall(msg)
        if ret:
            log.error("Failed to send message.")
            raise RuntimeError("Failed to send message.")
        ret = self.connection.recv(1)
        if ret != b'\x00':
            log.error("Failed to send %s command on port %d of %s." % (command, port_number, self.hostname))
            raise RuntimeError("Failed to send %s command on port %d of %s." % (command, port_number, self.hostname))

    def _close_connection(self):
        # Logout
        log.debug("Closing connection.")
        if self.password:
            log.debug("Attempting to logout.")
            ret = self.connection.sendall(b'\x7B')
            if ret:
                log.error("Failed to send message.")
                raise RuntimeError("Failed to send message.")
            ret = self.connection.recv(1)
            if ret != b'\x00':
                log.error("Failed to logout of %s." % self.hostname)
                raise RuntimeError("Failed to logout of %s." % self.hostname)
        self.connection.close()

    def _cleanup(self):
        self._close_connection()

    def _bombout(self):
        self._close_connection()

    @classmethod
    def accepts(cls, drivername):
        return False


class DevantechETH002(DevantechBase):
    port_count = 2

    @classmethod
    def accepts(cls, drivername):
        if drivername == "devantech_eth002":
            return True
        return False


class DevantechETH0621(DevantechBase):
    port_count = 2

    @classmethod
    def accepts(cls, drivername):
        if drivername == "devantech_eth0621":
            return True
        return False


class DevantechETH484(DevantechBase):
    port_count = 4

    @classmethod
    def accepts(cls, drivername):
        if drivername == "devantech_eth484":
            return True
        return False


class DevantechETH008(DevantechBase):
    port_count = 8

    @classmethod
    def accepts(cls, drivername):
        if drivername == "devantech_eth008":
            return True
        return False


class DevantechETH8020(DevantechBase):
    port_count = 20

    @classmethod
    def accepts(cls, drivername):
        if drivername == "devantech_eth8020":
            return True
        return False


class DevantechDSBase(PDUDriver):
    port_count = 0
    connection = None

    def __init__(self, hostname, settings):
        self.hostname = hostname
        self.settings = settings
        # the default hostname may vary. I.e. could be "ds378" for other modules.
        self.ip = settings.get("ip", "dS2824")
        self.port = settings.get("port", 17123)
        super().__init__()

    def connect(self):
        self.connection = socket.create_connection((self.ip, self.port))

    def port_interaction(self, command, port_number):
        self.connect()
        port_number = int(port_number)
        if port_number > self.port_count:
            log.error("There are only %d ports. Provide a port number lesser than %d." % (self.port_count, self.port_count))
            raise RuntimeError("There are only %d ports. Provide a port number lesser than %d." % (self.port_count, self.port_count))

        msg = 'SR {} '.format(port_number)
        if command in ["on", "off"]:
            msg += command
        else:
            log.error("Unknown command %s." % (command))
            return
        log.debug("Attempting control: %s port: %d hostname: %s." % (command, port_number, self.hostname))
        ret = self.connection.sendall(msg.encode('ascii'))
        if ret:
            log.error("Failed to send message.")
            raise RuntimeError("Failed to send message.")

    def _close_connection(self):
        # Logout
        log.debug("Closing connection.")
        self.connection.close()

    def _cleanup(self):
        self._close_connection()

    def _bombout(self):
        self._close_connection()

    @classmethod
    def accepts(cls, drivername):
        return False


class DevantechDS2824(DevantechDSBase):
    port_count = 8

    @classmethod
    def accepts(cls, drivername):
        # Some dS2824 modules apparently report themselves as `Module Type: dS378`
        if drivername in ["devantech_ds378", "devantech_ds2824"]:
            return True
        return False



#!/usr/bin/python3

#  Copyright 2017 Sjoerd Simons <sjoerd.simons@collabora.co.uk>
#
#  Based on PDUDriver:
#     Copyright 2013 Linaro Limited
#     Author Matt Hart <matthew.hart@linaro.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import logging
from pdudaemon.drivers.driver import PDUDriver
import serial

import os
log = logging.getLogger("pdud.drivers." + os.path.basename(__file__))


class DevantechusbBase(PDUDriver):
    connection = None
    port_count = 0
    supported = []

    def __init__(self, hostname, settings):
        self.hostname = hostname
        self.settings = settings
        self.device = settings.get("device", "/dev/ttyACM0")
        log.debug("device: %s" % self.device)

        super(DevantechusbBase, self).__init__()

    def port_interaction(self, command, port_number):
        port_number = int(port_number)
        if port_number > self.port_count or port_number < 1:
            err = "Port should be in the range 1 - %d" % (self.port_count)
            log.error(err)
            raise RuntimeError(err)

        if command == "on":
            byte = 0x64 + port_number
        elif command == "off":
            byte = 0x6e + port_number
        else:
            log.error("Unknown command %s." % (command))
            return

        s = serial.serial_for_url(self.device, 9600)
        s.write([byte])
        s.close()

    @classmethod
    def accepts(cls, drivername):
        return drivername in cls.supported


class DevantechUSB2(DevantechusbBase):
    port_count = 2
    supported = ["devantech_USB-RLY02",
                 "devantech_USB-RLY82"]


# Various 8 relay devices
class DevantechUSB8(DevantechusbBase):
    port_count = 8
    supported = ["devantech_USB-RLY08B",
                 "devantech_USB-RLY16",
                 "devantech_USB-RLY16L",
                 "devantech_USB-OPTO-RLY88",
                 "devantech_USB-OPTO-RLY816"]



#!/usr/bin/python3

#  Copyright 2013 Linaro Limited
#  Author Matt Hart <matthew.hart@linaro.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import asyncio
import logging
import time
import traceback
import pexpect
import concurrent.futures
from pdudaemon.drivers.driver import PDUDriver
import pdudaemon.drivers.strategies  # pylint: disable=W0611


class PDURunner:

    def __init__(self, config, hostname, retries):
        self.config = config
        self.hostname = hostname
        self.retries = retries
        self.logger = logging.getLogger("pdud.pdu.%s" % hostname)
        self.driver = self.driver_from_hostname(hostname)
        # use single-worker ThreadPoolExecutor to serialize execution
        self.executor = concurrent.futures.ThreadPoolExecutor(max_workers=1, thread_name_prefix=hostname)

    async def shutdown(self):
        self.executor.shutdown(wait=True, cancel_futures=True)

    def driver_from_hostname(self, hostname):
        drivername = self.config['driver']
        driver = PDUDriver.select(drivername)(hostname, self.config)
        return driver

    def do_job(self, port, request):
        self.logger.info("Processing job for PDU %s: (%s %s)", self.hostname, request, port)
        retries = self.retries
        while retries > 0:
            try:
                return self.driver.handle(request, port)
            except (OSError, pexpect.exceptions.EOF, Exception):  # pylint: disable=broad-except
                self.logger.warn(traceback.format_exc())
                self.logger.warn("Failed to execute job: {} {} (attempts left {})".format(port, request, retries - 1))
                if self.driver:
                    self.driver._bombout()  # pylint: disable=W0212,E1101
                time.sleep(5)
                retries -= 1
                continue
        return False

    async def do_job_async(self, port, request):
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(self.executor, self.do_job, port, request)



#!/usr/bin/python3

#  Copyright 2016 BayLibre, Inc.
#  Author Kevin Hilman
#
#  TODO:
#  - use pysnmp instead of snmpset command-line tool
#
#  Based on localcmdline.py
#  Copyright 2013 Linaro Limited
#  Author Matt Hart <matthew.hart@linaro.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import logging
import subprocess
from pdudaemon.drivers.localbase import LocalBase

import os
log = logging.getLogger("pdud.drivers." + os.path.basename(__file__))


class IP9258(LocalBase):

    def __init__(self, hostname, settings):
        self.hostname = hostname
        self.settings = settings

    @classmethod
    def accepts(cls, drivername):
        if drivername == "ip9258":
            return True
        return False

    def _port_interaction(self, command, port_number):
        port_number = int(port_number)
        power_oid = '1.3.6.1.4.1.92.58.2.%d.0' % (port_number)
        cmd_base = '/usr/bin/snmpset -v 1 -c public %s %s integer' \
                   % (self.hostname, power_oid)
        cmd = None

        if command == "on":
            cmd = cmd_base + ' %d > /dev/null' % (1)

        elif command == "off":
            cmd = cmd_base + ' %d > /dev/null' % (0)

        else:
            logging.debug("Unknown command!")

        if cmd:
            log.debug("running %s" % cmd)
            subprocess.call(cmd, shell=True)



#!/usr/bin/python3

#  Copyright 2013 Linaro Limited
#  Author Matt Hart <matthew.hart@linaro.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

from pdudaemon.drivers.acme import ACME  # pylint: disable=W0611
from pdudaemon.drivers.anelnetpwrctrl import AnelNETPwrCtrlHOME  # pylint: disable=W0611
from pdudaemon.drivers.anelnetpwrctrl import AnelNETPwrCtrlADV  # pylint: disable=W0611
from pdudaemon.drivers.anelnetpwrctrl import AnelNETPwrCtrlIO  # pylint: disable=W0611
from pdudaemon.drivers.anelnetpwrctrl import AnelNETPwrCtrlPRO  # pylint: disable=W0611
from pdudaemon.drivers.apc7900 import APC7900  # pylint: disable=W0611
from pdudaemon.drivers.apc7932 import APC7932  # pylint: disable=W0611
from pdudaemon.drivers.apc7952 import APC7952  # pylint: disable=W0611
from pdudaemon.drivers.apc9218 import APC9218  # pylint: disable=W0611
from pdudaemon.drivers.apc8959 import APC8959  # pylint: disable=W0611
from pdudaemon.drivers.apc9210 import APC9210  # pylint: disable=W0611
from pdudaemon.drivers.apc7920 import APC7920  # pylint: disable=W0611
from pdudaemon.drivers.apc7921 import APC7921  # pylint: disable=W0611
from pdudaemon.drivers.cleware import ClewareUsbSwitch4
from pdudaemon.drivers.ubiquity import Ubiquity3Port  # pylint: disable=W0611
from pdudaemon.drivers.ubiquity import Ubiquity6Port  # pylint: disable=W0611
from pdudaemon.drivers.ubiquity import Ubiquity8Port  # pylint: disable=W0611
from pdudaemon.drivers.localcmdline import LocalCmdline
from pdudaemon.drivers.ip9258 import IP9258
from pdudaemon.drivers.sainsmart import Sainsmart
from pdudaemon.drivers.devantech import DevantechETH002
from pdudaemon.drivers.devantech import DevantechETH0621
from pdudaemon.drivers.devantech import DevantechETH484
from pdudaemon.drivers.devantech import DevantechETH008
from pdudaemon.drivers.devantech import DevantechETH8020
from pdudaemon.drivers.devantech import DevantechDS2824
from pdudaemon.drivers.devantechusb import DevantechUSB2
from pdudaemon.drivers.devantechusb import DevantechUSB8
from pdudaemon.drivers.numatousb import NumatoUSB1
from pdudaemon.drivers.numatousb import NumatoUSB2
from pdudaemon.drivers.numatousb import NumatoUSB4
from pdudaemon.drivers.numatousb import NumatoUSB8
from pdudaemon.drivers.numatousb import NumatoUSB16
from pdudaemon.drivers.numatousb import NumatoUSB32
from pdudaemon.drivers.numatousb import NumatoUSB64
from pdudaemon.drivers.servertechpro2 import ServerTechPro2  # pylint: disable=W0611
from pdudaemon.drivers.synaccess import SynNetBooter
from pdudaemon.drivers.tasmota import SonoffS20Tasmota  # pylint: disable=W0611
from pdudaemon.drivers.tasmota import BrennenstuhlWSPL01Tasmota  # pylint: disable=W0611
from pdudaemon.drivers.egpms import EgPMS
from pdudaemon.drivers.ykush import YkushXS
from pdudaemon.drivers.ykush import Ykush
from pdudaemon.drivers.snmp import SNMP
from pdudaemon.drivers.energenieusb import EnerGenieUSB
from pdudaemon.drivers.bcu import BCU
from pdudaemon.drivers.vusbhid import VUSBHID
from pdudaemon.drivers.tplink import TPLink
from pdudaemon.drivers.ip9850 import ip9850
from pdudaemon.drivers.intellinet import Intellinet
from pdudaemon.drivers.esphome import ESPHomeHTTP
from pdudaemon.drivers.servo import Servo



#!/usr/bin/python3

#  Copyright 2015 Alexander Couzens <lynxis@fe80.eu>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import logging
from paramiko import SSHClient
from paramiko.ssh_exception import SSHException
from paramiko import RejectPolicy, WarningPolicy
from pdudaemon.drivers.driver import PDUDriver
import os
log = logging.getLogger("pdud.drivers." + os.path.basename(__file__))


class UbiquityBase(PDUDriver):
    client = None
    # overwrite power_count
    port_count = 0

    def __init__(self, hostname, settings):
        self.hostname = hostname
        self.settings = settings

        self.sshport = 22
        self.username = "ubnt"
        self.password = "ubnt"

        # verify ssh hostkey? unknown hostkey will make this job fail
        self.verify_hostkey = True

        if "sshport" in settings:
            self.sshport = settings["sshport"]
        if "username" in settings:
            self.username = settings["username"]
        if "password" in settings:
            self.password = settings["password"]
        if "verify_hostkey" in settings:
            self.verify_hostkey = settings["verify_hostkey"]

        super(UbiquityBase, self).__init__()

    def connect(self):
        log.info("Connecting to Ubiquity mfi %s@%s:%d",
                 self.username, self.hostname, self.sshport)
        self.client = SSHClient()
        self.client.load_system_host_keys()

        if self.verify_hostkey:
            self.client.set_missing_host_key_policy(RejectPolicy())
        else:
            self.client.set_missing_host_key_policy(WarningPolicy())

        self.client.connect(hostname=self.hostname, port=self.sshport,
                            username=self.username, password=self.password)

    def port_interaction(self, command, port_number):
        port_number = int(port_number)
        log.debug("Running port_interaction from UbiquityBase")
        self.connect()
        if port_number > self.port_count:
            raise RuntimeError("We only have ports 1 - %d. %d > maxPorts (%d)"
                               % self.port_count, port_number, self.port_count)

        if command == "on":
            command = "sh -c 'echo 1 > /proc/power/relay%d'" % port_number
        elif command == "off":
            command = "sh -c 'echo 0 > /proc/power/relay%d'" % port_number

        try:
            stdin, stdout, stderr = self.client.exec_command(command,
                                                             bufsize=-1,
                                                             timeout=3)
            stdin.close()
        except SSHException:
            pass

    def _cleanup(self):
        self.client.close()

    def _bombout(self):
        self.client.close()

    @classmethod
    def accepts(cls, drivername):
        return False


class Ubiquity3Port(UbiquityBase):
    port_count = 3

    @classmethod
    def accepts(cls, drivername):
        if drivername == "ubntmfi3port":
            return True
        return False


class Ubiquity6Port(UbiquityBase):
    port_count = 6

    @classmethod
    def accepts(cls, drivername):
        if drivername == "ubntmfi6port":
            return True
        return False


class Ubiquity8Port(UbiquityBase):
    port_count = 8

    @classmethod
    def accepts(cls, drivername):
        if drivername == "ubntmfi8port":
            return True
        return False



#!/usr/bin/python3

#  Copyright 2019 Martyn Welch <martyn.welch@collabora.com>
#
#  Based on devantechusb:
#     Copyright 2017 Sjoerd Simons
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import logging
from pdudaemon.drivers.driver import PDUDriver
import usb.core
import usb.util

import os
log = logging.getLogger("pdud.drivers." + os.path.basename(__file__))


class EnerGenieUSB(PDUDriver):
    supported = ["EG-PMS",
                 "EG-PM2"]

    def __init__(self, hostname, settings):
        self.hostname = hostname
        self.settings = settings
        self.device = settings.get("device", "")
        log.debug("device: %s" % self.device)

    def port_interaction(self, command, port_number):

        devices = self.connect()

        if len(devices) == 0:
            err = "No device found"
            log.error(err)
            raise RuntimeError(err)

        dev = None
        for d in devices:
            if self.getid(d) == self.device:
                dev = d
                break
        if dev is None:
            err = "Device with id {} not found".format(self.device)
            log.error(err)
            raise RuntimeError(err)

        port_number = int(port_number)
        if port_number > 4 or port_number < 1:
            err = "Port should be in the range 1 - 4"
            log.error(err)
            raise RuntimeError(err)

        if command == "on":
            self.switchon(dev, port_number)
        elif command == "off":
            self.switchoff(dev, port_number)
        else:
            log.error("Unknown command %s." % (command))
            return

    @classmethod
    def accepts(cls, drivername):
        return drivername in cls.supported

    # Following (modified) routines taken from pysispm:
    #
    # Copyright (c) 2016, Heinrich Schuchardt <xypron.glpk@gmx.de>
    # All rights reserved.
    #
    # Redistribution and use in source and binary forms, with or without
    # modification, are permitted provided that the following conditions are met:
    #
    #     * Redistributions of source code must retain the above copyright
    #       notice, this list of conditions and the following disclaimer.
    #
    #     * Redistributions in binary form must reproduce the above copyright
    #       notice, this list of conditions and the following disclaimer in the
    #       documentation and/or other materials provided with the distribution.
    #
    # THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
    # AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
    # IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
    # ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
    # DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
    # (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
    # LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
    # ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    # (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
    # SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

    def connect(self):
        """
        Returns the list of compatible devices.
        @return: device list
        """
        ret = list()
        ret += list(usb.core.find(find_all=True, idVendor=0x04b4, idProduct=0xfd10))
        ret += list(usb.core.find(find_all=True, idVendor=0x04b4, idProduct=0xfd11))
        ret += list(usb.core.find(find_all=True, idVendor=0x04b4, idProduct=0xfd12))
        ret += list(usb.core.find(find_all=True, idVendor=0x04b4, idProduct=0xfd13))
        ret += list(usb.core.find(find_all=True, idVendor=0x04b4, idProduct=0xfd15))
        return ret

    def getid(self, dev):
        """
        Gets the id of a device.
        @return: id
        """
        buf = bytes([0x00, 0x00, 0x00, 0x00, 0x00])
        id = dev.ctrl_transfer(0xa1, 0x01, 0x0301, 0, buf, 500)
        if (len(id) == 0):
            return None
        ret = ''
        sep = ''
        for x in id:
            ret += sep
            ret += format(x, '02x')
            sep = ':'
        return ret

    def switchoff(self, dev, i):
        """
        Switches outlet i of the device off.
        @param dev: device
        @param i: outlet
        """
        buf = bytes([3 * i, 0x00, 0x00, 0x00, 0x00])
        buf = dev.ctrl_transfer(0x21, 0x09, 0x0300 + 3 * i, 0, buf, 500)

    def switchon(self, dev, i):
        """
        Switches outlet i of the device on.
        @param dev: device
        @param i: outlet
        """
        buf = bytes([3 * i, 0x03, 0x00, 0x00, 0x00])
        buf = dev.ctrl_transfer(0x21, 0x09, 0x0300 + 3 * i, 0, buf, 500)



#!/usr/bin/python3

#  Copyright 2020 NXP
#  Author Leonard Crestez <leonard.crestez@nxp.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

from pdudaemon.drivers.driver import PDUDriver
import os
import subprocess
import logging
log = logging.getLogger("pdud.drivers." + os.path.basename(__file__))


class BCU(PDUDriver):
    """PDUDriver implementation using NXP BCU

    The NXP BCU tool can be used for remote power control on some newer NXP
    development boards.

    See https://github.com/NXPmicro/bcu
    """

    @classmethod
    def accepts(cls, drivername):
        return drivername == "bcu"

    def __init__(self, hostname, settings):
        self.hostname = hostname
        self.settings = settings

        #: usb path (string "like 3-7.1")
        self.id = settings.get('id', None)
        #: board id (see `bcu lsboard`)
        self.board = settings.get('board', None)
        #: path to bcu executable (default assumes bcu is in $PATH)
        self.bcu_exe = settings.get('bcu_exe', 'bcu')
        #: gpio used for reset function (see `bcu lsgpio -gpio=$gpio`)
        self.reset_gpio = settings.get('reset_gpio', 'reset')
        #: if reset_gpio is active low (default True)
        self.reset_gpio_active_low = bool(int(
            settings.get('reset_gpio_active_low', '1')))

        #: bootmode to initialize (see `bcu lsbootmode -board=$board`)
        self.bootmode = settings.get('bootmode', None)

    def _run(self, subcmd):
        cmd = [self.bcu_exe]
        cmd += subcmd
        if self.id:
            cmd += ['-id=' + self.id]
        if self.board:
            cmd += ['-board=' + self.board]
        log.info('RUN: %s', ' '.join(cmd))
        return subprocess.run(cmd, check=True)

    def _init(self):
        return self._run(['init'] + [self.bootmode] if self.bootmode else [])

    def _set_gpio(self, gpio, value):
        return self._run(['set_gpio', gpio, value])

    def port_on(self, port_number):
        self._init()
        self._set_gpio(
            self.reset_gpio,
            '1' if self.reset_gpio_active_low else '0')

    def port_off(self, port_number):
        self._init()
        self._set_gpio(
            self.reset_gpio,
            '0' if self.reset_gpio_active_low else '1')



#!/usr/bin/python3
#
#  Copyright 2018 Remi Duraffort <remi.duraffort@linaro.org>
#                 Matt Hart <matt@mattface.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

from setuptools import setup

# grab metadata without importing the module
metadata = {}
with open("pdudaemon/__about__.py", encoding="utf-8") as fp:
    exec(fp.read(), metadata)

# Setup the package
setup(
    name='pdudaemon',
    version=metadata['__version__'],
    description=metadata['__description__'],
    author=metadata['__author__'],
    author_email='matt@mattface.org',
    license=metadata['__license__'],
    url=metadata['__url__'],
    python_requires=">=3.4",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Communications",
        "Topic :: Software Development :: Testing",
        "Topic :: System :: Networking",
    ],
    packages=['pdudaemon', 'pdudaemon.drivers'],
    entry_points={
        'console_scripts': [
            'pdudaemon = pdudaemon:main'
        ]
    },
    install_requires=[
        "requests",
        "pexpect",
        "systemd_python",
        "paramiko",
        "pyserial",
        "hidapi",
        "pysnmp",
        "pyasn1",
        "pyusb"
    ],
    extras_require={
        'test': [
            'pytest',
            'pytest-mock',
        ],
    },
    zip_safe=True
)



#!/usr/bin/python3

#  Copyright 2015 BayLibre SAS
#  Author Marc Titinger <mtitinger@baylibre.com>
#
#  Based on ACPBase:
#     Copyright 2013 Linaro Limited
#     Author Matt Hart <matthew.hart@linaro.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import logging
import pexpect
from pdudaemon.drivers.driver import PDUDriver
import os
log = logging.getLogger("pdud.drivers." + os.path.basename(__file__))

# SSH connection, assuming the id_rsa.pub key for the owner of pdudaemon-runner
# i.e. root, was added to the authorized_keys on the ACME device.
# This may be changed to a simple telnet connection in the future.


class ACMEBase(PDUDriver):
    connection = None

    def __init__(self, hostname, settings):
        self.hostname = hostname
        self.settings = settings
        self.username = settings.get("username", "root")
        self.exec_string = "/usr/bin/ssh %s@%s" % (self.username, hostname)
        super(ACMEBase, self).__init__()

    @classmethod
    def accepts(cls, drivername):  # pylint: disable=unused-argument
        return False

    def port_interaction(self, command, port_number):
        log.debug("Running port_interaction from ACMEBase")
        self.get_connection()
        self._port_interaction(command,  # pylint: disable=no-member
                               port_number)

    def get_connection(self):
        log.debug("Connecting to Baylibre ACME with: {}".format(self.exec_string))
        # only uncomment this line for FULL debug when developing
        # self.connection = pexpect.spawn(self.exec_string, logfile=sys.stdout)
        self.connection = pexpect.spawn(self.exec_string)
        self._pdu_login(self.username, "")

    def _cleanup(self):
        self._pdu_logout()  # pylint: disable=no-member

    def _bombout(self):
        log.debug("Bombing out of driver: {}".format(self.connection))
        self.connection.close(force=True)
        del self

    def _pdu_login(self, username, password):
        log.debug("attempting login with username %s, password %s",
                  username, password)
        index = self.connection.expect(['#', 'password', 'yes/no'])
        if index == 1:
            self.connection.send("%s\r" % password)
        elif index == 2:
            self.connection.send("yes\r")



#!/usr/bin/python3

#  Copyright 2013 Linaro Limited
#  Author Matt Hart <matthew.hart@linaro.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import logging
from subprocess import call
from pdudaemon.drivers.localbase import LocalBase
import os
log = logging.getLogger("pdud.drivers." + os.path.basename(__file__))


class LocalCmdline(LocalBase):

    def __init__(self, hostname, settings):
        self.hostname = hostname
        self.settings = settings
        self.cmd_on = settings.get("cmd_on", None)
        self.cmd_off = settings.get("cmd_off", None)

    @classmethod
    def accepts(cls, drivername):
        if drivername == "localcmdline":
            return True
        return False

    def _port_interaction(self, command, port_number):
        cmd = None
        if command == "on" and self.cmd_on:
            cmd = self.cmd_on % port_number
        elif command == "off" and self.cmd_off:
            cmd = self.cmd_off % port_number
        else:
            log.debug("Unknown command!")

        if cmd:
            log.debug("running %s" % cmd)
            call(cmd, shell=True)



#!/usr/bin/python3
#
#  Copyright 2018 Remi Duraffort <remi.duraffort@linaro.org>
#                 Matt Hart <matt@mattface.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import argparse
import asyncio
import contextlib
import json
import logging
from logging.handlers import WatchedFileHandler
import signal
import sys
import time

from pdudaemon.tcplistener import TCPListener
from pdudaemon.httplistener import HTTPListener
from pdudaemon.pdurunner import PDURunner
from pdudaemon.drivers.driver import PDUDriver


###########
# Constants
###########
CONFIGURATION_FILE = "/etc/pdudaemon/pdudaemon.conf"
logging_FORMAT = "%(asctime)s - %(name)-30s - %(levelname)s %(message)s"
logging_FORMAT_JOURNAL = "%(name)s.%(levelname)s %(message)s"
logging_FILE = "/var/log/pdudaemon.log"

##################
# Global logger
##################
logger = logging.getLogger('pdud')


def setup_logging(options, settings):
    logger = logging.getLogger("pdud")
    """
    Setup the log handler and the log level
    """
    if options.journal:
        from systemd.journal import JournalHandler
        handler = JournalHandler(SYSLOG_IDENTIFIER="pdudaemon")
        handler.setFormatter(logging.Formatter(logging_FORMAT_JOURNAL))
    elif options.logfile == "-" or not options.logfile:
        handler = logging.StreamHandler(sys.stdout)
        handler.setFormatter(logging.Formatter(logging_FORMAT))
    else:
        handler = WatchedFileHandler(options.logfile)
        handler.setFormatter(logging.Formatter(logging_FORMAT))

    logger.addHandler(handler)
    settings_level = settings.get('daemon', {}).get('logging_level', None)
    if settings_level:
        options.loglevel = settings_level.upper()
    else:
        options.loglevel = options.loglevel.upper()
    if options.loglevel == "DEBUG":
        logger.setLevel(logging.DEBUG)
    elif options.loglevel == "INFO":
        logger.setLevel(logging.INFO)
    elif options.loglevel == "WARNING":
        logger.setLevel(logging.WARNING)
    else:
        logger.setLevel(logging.ERROR)


class PDUDaemon:
    def __init__(self, options, settings):
        # Context
        self.runners = {}

        # Create the runners
        logger.info("Creating the runners")
        for hostname in settings["pdus"]:
            config = settings["pdus"][hostname]
            retries = config.get("retries", 5)
            self.runners[hostname] = PDURunner(config, hostname, retries)

        # Start the listener
        logger.info("Starting the listener")
        if options.listener:
            listener = options.listener
        else:
            listener = settings['daemon'].get('listener', 'tcp')
        if listener == 'tcp':
            self.listener = TCPListener(settings, self)
        elif listener == 'http':
            self.listener = HTTPListener(settings, self)
        else:
            logging.error("Unknown listener configured")

    async def start(self):
        await self.listener.start()

    async def shutdown(self):
        logger.info("Shutting down listener...")
        await self.listener.shutdown()
        logger.info("Shutting down runners...")
        for runner in self.runners.values():
            await runner.shutdown()
        logger.info("Stopping loop...")
        loop = asyncio.get_running_loop()
        loop.stop()


async def main_async():
    # Setup the parser
    parser = argparse.ArgumentParser()

    log = parser.add_argument_group("logging")
    log.add_argument("--journal", "-j", action="store_true", default=False,
                     help="Log to the journal")
    log.add_argument("--logfile", dest="logfile", action="store", type=str,
                     default="-", help="log file [%s]" % logging_FILE)
    log.add_argument("--loglevel", dest="loglevel", default="INFO",
                     choices=["DEBUG", "ERROR", "INFO", "WARN"],
                     type=str, help="logging level [INFO]")
    parser.add_argument("--conf", "-c", type=argparse.FileType("r"),
                        default=CONFIGURATION_FILE,
                        help="configuration file [%s]" % CONFIGURATION_FILE)
    parser.add_argument("--listener", type=str, help="PDUDaemon listener setting")
    conflict = parser.add_mutually_exclusive_group()
    conflict.add_argument("--alias", dest="alias", action="store", type=str)
    conflict.add_argument("--hostname", dest="drivehostname", action="store", type=str)
    drive = parser.add_argument_group("drive")
    drive.add_argument("--drive", action="store_true", default=False)
    drive.add_argument("--request", dest="driverequest", action="store", type=str)
    drive.add_argument("--retries", dest="driveretries", action="store", type=int, default=5)
    drive.add_argument("--port", dest="driveport", action="store", type=str)

    # Parse the command line
    options = parser.parse_args()

    # Read the configuration file
    try:
        settings = json.loads(options.conf.read())
    except Exception as exc:
        logging.error("Unable to read configuration file '%s': %s", options.conf.name, exc)
        return 1

    # Setup logging
    setup_logging(options, settings)

    # Get handle to the currently running loop
    loop = asyncio.get_running_loop()

    if options.drive:
        # Driving a PDU directly, dont start any Listeners

        if options.alias:
            # Using alias support, get all pdu info from alias
            alias_settings = settings["aliases"].get(options.alias, False)
            if not alias_settings:
                logging.error("Alias requested but not found")
                sys.exit(1)
            options.drivehostname = settings["aliases"][options.alias]["hostname"]
            options.driveport = settings["aliases"][options.alias]["port"]

        # Check that the requested PDU has config
        config = settings["pdus"].get(options.drivehostname, False)
        if not config:
            logging.error("No config section for hostname: {}".format(options.drivehostname))
            sys.exit(1)

        runner = PDURunner(config, options.drivehostname, options.driveretries)
        if options.driverequest == "reboot":
            result = await runner.do_job_async(options.driveport, "off")
            result = await runner.do_job_async(options.driveport, "on")
        else:
            result = await runner.do_job_async(options.driveport, options.driverequest)
        await runner.shutdown()
        loop.stop()
        return result

    # Create daemon
    logger.info('PDUDaemon starting up')
    daemon = PDUDaemon(options, settings)

    # Setup signal handling
    def cleanup_handler():
        logger.info("Signal received, shutting down...")
        asyncio.create_task(daemon.shutdown())

    loop.add_signal_handler(signal.SIGINT, cleanup_handler)
    loop.add_signal_handler(signal.SIGTERM, cleanup_handler)

    await daemon.start()


def main():
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main_async())
    try:
        loop.run_forever()
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    # execute only if run as a script
    result = main()
    sys.exit(result)



#!/usr/bin/python3

#  Copyright 2017 Matt Hart <matt@mattface.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import logging
from pdudaemon.drivers.apc7952 import APC7952
import os
log = logging.getLogger("pdud.drivers." + os.path.basename(__file__))


class APC7921(APC7952):

    @classmethod
    def accepts(cls, drivername):
        if drivername == "apc7921":
            return True
        return False

    def _port_interaction(self, command, port_number):
        # make sure in main menu here
        self._back_to_main()
        self.connection.send("\r")
        self.connection.expect("1- Device Manager")
        self.connection.expect("> ")
        log.debug("Entering Device Manager")
        self.connection.send("1\r")
        self.connection.expect("------- Device Manager")
        self.connection.send("2\r")
        self.connection.expect("1- Outlet Control/Configuration")
        self.connection.expect("> ")
        self.connection.send("1\r")
        self._enter_outlet(port_number, False)
        self.connection.expect("1- Control Outlet")
        self.connection.send("1\r")
        self.connection.expect("> ")
        if command == "on":
            self.connection.send("1\r")
            self.connection.expect("Immediate On")
            self._do_it()
        elif command == "off":
            self.connection.send("2\r")
            self.connection.expect("Immediate Off")
            self._do_it()
        else:
            log.debug("Unknown command!")



#!/usr/bin/python3

#  Copyright 2013 Linaro Limited
#  Author Matt Hart <matthew.hart@linaro.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import logging
from pdudaemon.drivers.apcbase import APCBase
import os
log = logging.getLogger("pdud.drivers." + os.path.basename(__file__))


class APC7952(APCBase):

    @classmethod
    def accepts(cls, drivername):
        if drivername == "apc7952":
            return True
        return False

    def _pdu_logout(self):
        self._back_to_main()
        log.debug("Logging out")
        self.connection.send("4\r")

    def _back_to_main(self):
        log.debug("Returning to main menu")
        self.connection.send("\r")
        self.connection.expect('>')
        for _ in range(1, 20):
            self.connection.send("\x1B")
            self.connection.send("\r")
            res = self.connection.expect(["4- Logout", "> "])
            if res == 0:
                log.debug("Back at main menu")
                break

    def _enter_outlet(self, outlet, enter_needed=True):
        outlet = "%s" % outlet
        log.debug("Attempting to enter outlet %s", outlet)
        if enter_needed:
            self.connection.expect("Press <ENTER> to continue...")
        log.debug("Sending enter")
        self.connection.send("\r")
        self.connection.expect("> ")
        log.debug("Sending outlet number")
        self.connection.send(outlet)
        self.connection.send("\r")
        log.debug("Finished entering outlet")

    def _port_interaction(self, command, port_number):
        # make sure in main menu here
        self._back_to_main()
        self.connection.send("\r")
        self.connection.expect("1- Device Manager")
        self.connection.expect("> ")
        log.debug("Entering Device Manager")
        self.connection.send("1\r")
        self.connection.expect("------- Device Manager")
        self.connection.send("2\r")
        self.connection.expect("1- Outlet Control/Configuration")
        self.connection.expect("> ")
        self.connection.send("1\r")
        self._enter_outlet(port_number, False)
        self.connection.expect("> ")
        self.connection.send("1\r")
        res = self.connection.expect(["> ", "Press <ENTER> to continue..."])
        if res == 1:
            log.debug("Stupid paging thingmy detected, pressing enter")
            self.connection.send("\r")
        self.connection.send("\r")
        if command == "on":
            self.connection.send("1\r")
            self.connection.expect("Immediate On")
            self._do_it()
        elif command == "off":
            self.connection.send("2\r")
            self.connection.expect("Immediate Off")
            self._do_it()
        else:
            log.debug("Unknown command!")

    def _do_it(self):
        self.connection.expect("Enter 'YES' to continue or "
                               "<ENTER> to cancel :")
        self.connection.send("YES\r")
        self.connection.expect("Press <ENTER> to continue...")
        self.connection.send("\r")



#!/usr/bin/python3

#  Copyright 2017 Sjoerd Simons Schulz <sjoerd.simons@collabora.co.uk>
#
#  Based on PDUDriver:
#     Copyright 2013 Linaro Limited
#     Author Matt Hart <matthew.hart@linaro.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import logging
from pdudaemon.drivers.driver import PDUDriver
import socket
import sys
from array import array

import os
log = logging.getLogger("pdud.drivers." + os.path.basename(__file__))


class EgPMS(PDUDriver):
    port_count = 4

    def __init__(self, hostname, settings):
        self.hostname = hostname
        self.settings = settings
        self.ip = settings["ip"]
        self.password = array('B', settings["password"].encode("utf-8") + 8 * b' ')[:8]
        self.challenge = None

    def authresponse(self, part):
        response = self.challenge[part] ^ self.password[part + 2]
        response *= self.password[part]
        response ^= self.password[part + 4] << 8 | self.password[part + 6]
        response ^= self.challenge[part + 2]

        r = array('B', [response & 0xff, (response >> 8) & 0xff])
        return r

    def encode_state(self, status):
        state = status ^ self.challenge[2]
        state += self.challenge[3]
        state ^= self.password[0]
        state += self.password[1]

        return state & 0xff

    def decode_state(self, state):
        state = state - self.password[1]
        state ^= self.password[0]
        state -= self.challenge[3]
        state ^= self.challenge[2]

        return state & 0xff

    def dump_status(self, status):
        status.reverse()
        for b in status:
            s = self.decode_state(b)
            log.debug("on: %d (raw: %x)" % (s >> 4 == 1, s))

    def connect(self):
        self.socket = socket.create_connection((self.ip, 5000))
        self.socket.send(b'\x11')
        self.challenge = array('B', self.socket.recv(4))

        self.socket.send(self.authresponse(0) + self.authresponse(1))
        status = array('B', self.socket.recv(self.port_count))
        log.debug("Connected")
        self.dump_status(status)

    def disconnect(self):
        # From egctl:
        # Empirically found way to close session w/o 4 second timeout on
        # the device side is to send some invalid sequence. This helps
        # to avoid a hiccup on subsequent run of the utility.
        #
        # Other protocol documents explain that after the current settings the
        # device waits for a schedule update for a while (if any) but will go
        # back to the start state if it doesn't make sense.
        self.socket.send(b'\x11')
        self.socket.close()

    def port_interaction(self, command, port_number):
        SWITCH_ON = 0x1
        SWITCH_OFF = 0x2
        DONT_SWITCH = 0x4
        port_number = int(port_number)
        if port_number > self.port_count or port_number < 1:
            err = "Port should be in the range 1 - %d" % (self.port_count)
            log.error(err)
            raise RuntimeError(err)

        if command == "on":
            on = True
        elif command == "off":
            on = False
        else:
            log.error("Unknown command %s." % (command))
            return

        self.connect()
        log.debug("Attempting control: {} port: {} hostname: {}.".format(command, port_number, self.hostname))
        update = array('B')
        for s in range(1, self.port_count + 1):
            if s != port_number:
                update.append(self.encode_state(DONT_SWITCH))
            else:
                update.append(
                    self.encode_state(SWITCH_ON if on else SWITCH_OFF))

        # States get send in reverse port order bytewise
        update.reverse()

        self.socket.send(update)
        status = array('B', self.socket.recv(self.port_count))
        self.disconnect()

        log.debug("Updated")
        self.dump_status(status)

    @classmethod
    def accepts(cls, drivername):
        if drivername == 'egpms':
            return True
        return False



#!/usr/bin/python3

#  Copyright 2018 Matt Hart <matt@mattface.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import logging
from pysnmp.hlapi import setCmd, SnmpEngine, UsmUserData, UdpTransportTarget, ContextData, CommunityData, ObjectType, ObjectIdentity
import pysnmp.hlapi as pysnmp_api
from pdudaemon.drivers.driver import PDUDriver, FailedRequestException, UnknownCommandException
import os
log = logging.getLogger("pdud.drivers." + os.path.basename(__file__))


class SNMP(PDUDriver):

    def __init__(self, hostname, settings):
        self.hostname = hostname
        self.version = settings['driver']
        self.mib = settings['mib']
        self.authpass = settings.get('authpassphrase', None)
        self.privpass = settings.get('privpassphrase', None)
        self.community = settings.get('community', None)
        self.controlpoint = settings['controlpoint']
        self.username = settings.get('username', None)
        self.onsetting = settings['onsetting']
        self.offsetting = settings['offsetting']
        self.inside_number = settings.get('inside_number', None)
        self.static_ending = settings.get('static_ending', None)
        self.auth_protocol = settings.get('auth_protocol', None)
        self.priv_protocol = settings.get('priv_protocol', None)
        super(SNMP, self).__init__()

    @classmethod
    def accepts(cls, drivername):
        if drivername in ['snmpv3', 'snmpv1']:
            return True
        return False

    def validate(self):
        pass

    def port_interaction(self, command, port_number):
        if command == "on":
            set_bit = self.onsetting
        elif command == "off":
            set_bit = self.offsetting
        else:
            raise UnknownCommandException("Unknown command %s." % (command))

        transport = UdpTransportTarget((self.hostname, 161))

        if self.inside_number:
            # It is possible to pass 2 or 3 snmp argument values
            filled_controlpoint = self.controlpoint.replace('*', str(port_number))
            indexed_object_list = [self.mib, filled_controlpoint]
            # If there is a key static_ending available, add a static ending value
            if self.static_ending is not None:
                indexed_object_list.append(int(self.static_ending))
        else:
            indexed_object_list = [self.mib, self.controlpoint, port_number]

        objecttype = ObjectType(
            ObjectIdentity(*indexed_object_list).addAsn1MibSource(
                'http://mibs.snmplabs.com/asn1/@mib@'), int(set_bit))

        if self.version == 'snmpv3':
            if not self.username:
                raise FailedRequestException("No username set for snmpv3")

            protocols = {}
            if self.auth_protocol is not None:
                a_protocol = getattr(pysnmp_api, self.auth_protocol)
                protocols['authProtocol'] = a_protocol

            if self.priv_protocol is not None:
                p_protocol = getattr(pysnmp_api, self.priv_protocol)
                protocols['privProtocol'] = p_protocol

            userdata = UsmUserData(self.username, self.authpass, self.privpass, **protocols)
            errorIndication, errorStatus, errorIndex, varBinds = next(
                setCmd(SnmpEngine(),
                       userdata,
                       transport,
                       ContextData(),
                       objecttype)
            )
        elif self.version == 'snmpv1':
            if not self.community:
                raise FailedRequestException("No community set for snmpv1")
            errorIndication, errorStatus, errorIndex, varBinds = next(
                setCmd(SnmpEngine(),
                       CommunityData(self.community),
                       transport,
                       ContextData(),
                       objecttype)
            )
        else:
            raise FailedRequestException("Unknown snmp version")

        if errorIndication:
            raise FailedRequestException(errorIndication)
        elif errorStatus:
            raise FailedRequestException(errorStatus)
        else:
            for varBind in varBinds:
                log.debug(' = '.join([x.prettyPrint() for x in varBind]))
            return True



#  Copyright 2020 Leonard Crestez <cdleonard@gmail.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import logging
from pdudaemon.drivers.driver import PDUDriver
import serial

import os
log = logging.getLogger("pdud.drivers." + os.path.basename(__file__))


class NumatoUSB(PDUDriver):
    """Supports various numato USB relay modules

    Comes in many variants with similar protocols

    Protocol documentation link:
    https://numato.com/docs/16-channel-usb-relay-module/#the-command-set-23

    There are slight differences for 32 and 64-channel modules
    """
    def __init__(self, hostname, settings):
        super(NumatoUSB, self).__init__()
        self.hostname = hostname
        self.settings = settings
        self.device = settings.get("device", "/dev/ttyACM0")
        log.debug("device: %s" % self.device)
        self.serial_port = None

    def _read_reply(self):
        msg = self.serial_port.read_until(b'>').decode('utf8')
        x = msg.find('\n\r')
        return msg[x + 2:-3]

    def _open(self):
        self.serial_port = serial.serial_for_url(self.device, timeout=1)
        log.debug("open: %s for %s", self.serial_port.port, self.device)
        # resync prompt
        self.serial_port.reset_input_buffer()
        self.serial_port.write(b'\r')
        self.serial_port.read_until(b'>')
        # show relay module info
        self.serial_port.write(b'ver\r')
        log.info("ver: %s", self._read_reply())
        self.serial_port.write(b'id get\r')
        log.info("id: %s", self._read_reply())

    def port_interaction(self, command, port_number):
        port_number = int(port_number)
        if port_number > self.port_count or port_number < 1:
            err = "Port should be in the range 1 - %d" % (self.port_count)
            raise RuntimeError(err)

        if command == "on":
            msg = 'relay on %s\r' % (self.format_portid(port_number - 1),)
        elif command == "off":
            msg = 'relay off %s\r' % (self.format_portid(port_number - 1),)
        else:
            raise RuntimeError("Unknown command %s" % (command))

        try:
            if not self.serial_port:
                self._open()
            log.debug("write %r on %s", msg, self.serial_port.port)
            self.serial_port.write(msg.encode('utf8'))
            self.serial_port.read_until(b'>')
        except (serial.serialutil.SerialException, OSError) as e:
            log.error(e)
            self.serial_port = None
            raise

    def format_portid(self, portid):
        # See https://numato.com/docs/16-channel-usb-relay-module/#the-command-set-23
        return '%x' % portid

    @classmethod
    def accepts(cls, drivername):
        if not getattr(cls, 'port_count', None):
            return False
        return drivername == "NumatoUSB%d" % cls.port_count


class NumatoUSB1(NumatoUSB):
    port_count = 1


class NumatoUSB2(NumatoUSB):
    port_count = 2


class NumatoUSB4(NumatoUSB):
    port_count = 4


class NumatoUSB8(NumatoUSB):
    port_count = 8


class NumatoUSB16(NumatoUSB):
    port_count = 16


class NumatoUSB32(NumatoUSB):
    port_count = 32

    def format_portid(self, portid):
        # See https://numato.com/docs/32-channel-usb-relay-module/#the-command-set-23
        if portid < 10:
            return '%d' % portid
        else:
            return chr(ord('A') + portid - 10)


class NumatoUSB64(NumatoUSB):
    port_count = 64

    def format_portid(self, portid):
        # See https://numato.com/docs/64-channel-usb-relay-module-user-guide/#the-commands-set-4
        return '%02d' % portid



#!/usr/bin/python3

#  Copyright 2020 Arm Limited
#  Author Malcolm Brooks <malcolm.brooks@arm.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import json
import logging
import os

import requests

from pdudaemon.drivers.localbase import LocalBase

log = logging.getLogger("pdud.drivers." + os.path.basename(__file__))


class ServerTechPro2(LocalBase):

    def __init__(self, hostname, settings):
        self.hostname = hostname
        self.settings = settings
        self.username = settings.get("username", "admin")
        self.password = settings.get("password", "admin")
        self.insecure = bool(settings.get("insecure", False))

        protocol = settings.get("protocol", "http")
        server_address = settings.get("ip", self.hostname)
        self.url_base = "{}://{}/jaws/control/outlets/".format(
            protocol, server_address)

    @classmethod
    def accepts(cls, drivername):
        if drivername == "servertechpro2":
            return True
        return False

    def _port_interaction(self, command, port_number):
        log.debug("Attempting command: '{}' port: '{}'".format(command, port_number))
        if command in ("on", "off"):
            if self.insecure:
                requests.packages.urllib3.disable_warnings()
            url = "{}{}".format(self.url_base, port_number)
            data = {"control_action": command}
            r = requests.patch(
                url=url,
                data=json.dumps(data),
                auth=(self.username, self.password),
                headers={'Content-Type': 'application/json'},
                verify=not self.insecure
            )
            r.raise_for_status()
            log.debug("Done")
        else:
            log.debug("Unknown command!")
